﻿using Org.BouncyCastle.Math;
using System;
using System.Collections;
using System.Collections.Generic;
using System.Linq;
using System.Security.Cryptography;
using System.Text;
using System.Threading.Tasks;

namespace RSA
{
    /// <summary>
    /// RSA加密解密，支持公钥加密私钥解密和私钥加密公钥解密
    /// </summary>
    public class RSACryption
    {
        /// <summary>
        /// 密钥大小
        /// </summary>
        static int size = 1024;

        RSACryptoServiceProvider provider;

        /// <summary>
        /// 构造方法
        /// </summary>
        /// <param name="privateKey">私钥</param>
        public RSACryption(string privateKey)
        {
            provider = new RSACryptoServiceProvider();
            provider.FromXmlString(privateKey);
        }

        /// <summary>
        /// 获取密钥对
        /// </summary>
        /// <param name="privateKey">私钥</param>
        /// <param name="publicKey">公钥</param>
        /// <remarks>author:lorne date:2015-11-02</remarks>
        public static void GetKey(out string privateKey, out string publicKey)
        {
            RSACryptoServiceProvider provider = new RSACryptoServiceProvider();
            privateKey = provider.ToXmlString(true);
            publicKey = provider.ToXmlString(false);
        }

        /// <summary>
        /// 公钥加密
        /// </summary>
        /// <param name="content">待加密内容</param>
        /// <returns>加密结果</returns>
        /// <remarks>author:lorne date:2015-11-02</remarks>
        public string EncryptByPublicKey(string content)
        {
            byte[] source = Encoding.UTF32.GetBytes(content);
            int encrySize = size / 8;
            int blockSize = encrySize - 42;
            int dataLength = source.Length;
            int iterations = dataLength / blockSize;

            StringBuilder stringBuilder = new StringBuilder();
            for (int i = 0; i <= iterations; i++)
            {
                byte[] tempBytes = new byte[(dataLength - blockSize * i > blockSize) ? blockSize : dataLength - blockSize * i];
                Buffer.BlockCopy(source, blockSize * i, tempBytes, 0, tempBytes.Length);
                byte[] encryptedBytes = provider.Encrypt(tempBytes, true);
                Array.Reverse(encryptedBytes);
                stringBuilder.Append(Convert.ToBase64String(encryptedBytes));
            }

            return stringBuilder.ToString();
        }

        /// <summary>
        /// 私钥加密
        /// </summary>
        /// <param name="content">待加密内容</param>
        /// <returns>加密结果</returns>
        /// <remarks>author:lorne date:2015-11-02</remarks>
        public string EncryptByPrivateKey(string content)
        {
            var paramsters = provider.ExportParameters(true);
            BigInteger d = new BigInteger(1, paramsters.D);
            BigInteger n = new BigInteger(1, paramsters.Modulus);

            int encrySize = provider.KeySize / 8;
            byte[] source = Encoding.UTF32.GetBytes(content);
            int dataLength = source.Length;
            int iterations = Convert.ToInt32(Math.Ceiling(Convert.ToDouble(dataLength) / encrySize));

            StringBuilder stringBuilder = new StringBuilder();
            for (int i = 0; i < iterations; i++)
            {
                byte[] context = new byte[(dataLength - encrySize * i > encrySize) ? encrySize : dataLength - encrySize * i];
                Buffer.BlockCopy(source, encrySize * i, context, 0, context.Length);

                BigInteger biText = new BigInteger(1, context);
                BigInteger biEnText = biText.ModPow(d, n);

                byte[] b = biEnText.ToByteArrayUnsigned();
                stringBuilder.Append(Convert.ToBase64String(b));
            }
            return stringBuilder.ToString();
        }

        /// <summary>
        /// 私钥解密
        /// </summary>
        /// <param name="content">待解密内容</param>
        /// <returns>解密结果</returns>
        /// <remarks>author:lorne date:2015-11-02</remarks>
        public string DecryptByPrivateKey(string content)
        {
            int encrySize = provider.KeySize;
            int base64BlockSize = ((encrySize / 8) % 3 != 0) ? (((encrySize / 8) / 3) * 4) + 4 : ((encrySize / 8) / 3) * 4;
            int iterations = content.Length / base64BlockSize;

            ArrayList arrayList = new ArrayList();
            for (int i = 0; i < iterations; i++)
            {
                byte[] encryptedBytes = Convert.FromBase64String(content.Substring(base64BlockSize * i, base64BlockSize));
                Array.Reverse(encryptedBytes);
                arrayList.AddRange(provider.Decrypt(encryptedBytes, true));
            }

            return Encoding.UTF32.GetString(arrayList.ToArray(Type.GetType("System.Byte")) as byte[]);
        }

        /// <summary>
        /// 公钥解密
        /// </summary>
        /// <param name="content">待解密内容</param>
        /// <returns>解密结果</returns>
        /// <remarks>author:lorne date:2015-11-02</remarks>
        public string DecryptByPublicKey(string content)
        {
            var paramsters = provider.ExportParameters(false);
            BigInteger e = new BigInteger(1, paramsters.Exponent);
            BigInteger n = new BigInteger(1, paramsters.Modulus);

            int size = provider.KeySize;
            int base64BlockSize = ((size / 8) % 3 != 0) ? (((size / 8) / 3) * 4) + 4 : ((size / 8) / 3) * 4;
            int iterations = Convert.ToInt32(Math.Ceiling(Convert.ToDouble(content.Length) / base64BlockSize));

            ArrayList arrayList = new ArrayList();
            for (int i = 0; i < iterations; i++)
            {
                int length = base64BlockSize;
                if (content.Length - base64BlockSize * (i + 1) < 0)
                {
                    length = content.Length - base64BlockSize * i;
                }
                byte[] context = Convert.FromBase64String(content.Substring(base64BlockSize * i, length));

                BigInteger biText = new BigInteger(1, context);
                BigInteger biEnText = biText.ModPow(e, n);

                byte[] b = biEnText.ToByteArrayUnsigned();
                arrayList.AddRange(b);
            }

            return Encoding.UTF32.GetString(arrayList.ToArray(Type.GetType("System.Byte")) as byte[]);
        }
    }
}
